/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is WebMail Notifier.
 *
 * The Initial Developer of the Original Code is
 * Byungwook Kang.
 * Portions created by the Initial Developer are Copyright (C) 2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

/***********************************************************
constants
***********************************************************/

const nsIWebMailNotifier = Components.interfaces.nsIWebMailNotifier;

// UUID uniquely identifying our component
const CLASS_ID = Components.ID("{e3ab0980-b4e1-11db-abbd-0800200c9a66}");

// description
const CLASS_NAME = "WebMail Notifier Service";

// textual unique identifier
const CONTRACT_ID = "@mozilla.org/WebMailNotifier;1";

const ST_PRE         = 0;
const ST_PRE_RES     = 1;
const ST_LOGIN       = 10;
const ST_LOGIN_RES   = 11;
const ST_DATA        = 20;
const ST_DATA_RES    = 21;

const LOGIN_URL      = 0;
const LOGIN_ID       = 1;
const LOGIN_PW       = 2;
const LOGIN_DATA     = 3;
const LOGIN_POST     = 1;

const formSubmitURL = "user";

const dout=Components.utils.reportError;

var enableDebug=false;
/***********************************************************
class definition
***********************************************************/

//class constructor
function WebMailNotifier() {
  //preference observer
  var prefService = Components.classes["@mozilla.org/preferences-service;1"]
                              .getService(Components.interfaces.nsIPrefService);
  this.prefBranch = prefService.getBranch("wm-notifier.");
  this.prefBranch.QueryInterface(Components.interfaces.nsIPrefBranch2);
  this.prefBranch.addObserver("", this, false);
  this.timerDelay=this.prefBranch.getIntPref("updateInterval")*60000;
  this.newMailsOnly=this.prefBranch.getBoolPref("newMailsOnly");
  enableDebug=this.prefBranch.getIntPref("debug");

  try{
    this.passwordManager = Components.classes["@mozilla.org/passwordmanager;1"]
                                  .getService(Components.interfaces.nsIPasswordManager);
  }catch(e){}
  try{
    this.loginManager = Components.classes["@mozilla.org/login-manager;1"]
                           .getService(Components.interfaces.nsILoginManager);
  }catch(e){}

  var str="chrome://wm-notifier/options";
  var str2="autoLogin";
  this.autoLogin=this.getPassword(str,str2);
  if(this.autoLogin==null){
    this.autoLogin=1;
    this.setPassword(str,str2,"1");
  }else this.autoLogin=parseInt(this.autoLogin);

  var dir = Components.stack.toString();
  dir=dir.match(/:: (.+\/).+?/);
  dir=dir[1];
  var hosts=this.prefBranch.getIntPref("hosts");
  try{ this.prefBranch.clearUserPref("hosts"); }catch(e){}
  var defaultHosts=this.prefBranch.getIntPref("hosts");
  var loader =  Components.classes["@mozilla.org/moz/jssubscript-loader;1"]
                  .getService(Components.interfaces.mozIJSSubScriptLoader);
  this.loaded=[];
  for(var i=0;i<defaultHosts;i++){
    try{
      loader.loadSubScript(dir+"webmail"+i+".js");
      this.loaded[i]=true;
     }catch(e){}
  }

  var ios = Components.classes["@mozilla.org/network/io-service;1"]
                      .getService(Components.interfaces.nsIIOService);
  var fileHandler = ios.getProtocolHandler("file")
                       .QueryInterface(Components.interfaces.nsIFileProtocolHandler);
  var ar=this.getScriptList();
  var e = ar.enumerate();
  while(e.hasMoreElements())
  {
    var t = e.getNext().QueryInterface(Components.interfaces.nsIFile);
    t=fileHandler.getURLSpecFromFile(t);
    try{
      loader.loadSubScript(t);
    }catch(e){
dout(e);
    }
  }
  for(var i=this.loaded.length;i<hosts;i++){//check load state of user script host
    try{
      eval("initHandler"+i);
      this.loaded[i]=true;
    }catch(e){
dout(e);
    };
  }
  this.prefBranch.setIntPref("hosts",this.loaded.length);

  this.initHandlers(false);
  var observerService = Components.classes["@mozilla.org/observer-service;1"]
                                .getService(Components.interfaces.nsIObserverService);
  observerService.addObserver(this, "http-on-modify-request", false);
  observerService.addObserver(this, "http-on-examine-response", false);
}

// class definition
WebMailNotifier.prototype = {
  // define the function we want to expose in our interface
  listeners : new Array(),
  timer : null,
  countTotal : 0,
  QueryInterface: function(aIID)
  {
    if (!aIID.equals(nsIWebMailNotifier) &&
        !aIID.equals(Components.interfaces.nsISupports))
      throw Components.results.NS_ERROR_NO_INTERFACE;
    return this;
  }
}
WebMailNotifier.prototype.getScriptList = function() {
  var array = Components.classes["@mozilla.org/array;1"]
                        .createInstance(Components.interfaces.nsIMutableArray);
  var file = Components.classes["@mozilla.org/file/directory_service;1"]
                     .getService(Components.interfaces.nsIProperties)
                     .get("ProfD", Components.interfaces.nsIFile);
  file.append("wmn");
  if(file.exists()){
    var ios = Components.classes["@mozilla.org/network/io-service;1"]
                        .getService(Components.interfaces.nsIIOService);
    var entries = file.directoryEntries;
    while(entries.hasMoreElements())
    {
      var t = entries.getNext();
      t.QueryInterface(Components.interfaces.nsIFile);
      var s=t.path;
      if(s.match(/.js$/))array.appendElement(t,false);
    }
  }
  return array;
}
WebMailNotifier.prototype.initHandlers = function(aCheckNow) {
  var hosts=this.prefBranch.getIntPref("hosts");
  this.webMailHandlers=[];
  var str="chrome://wm-notifier/webmail";
  if(this.passwordManager){
    var e = this.passwordManager.enumerator;
    while (e.hasMoreElements()) {
      try {
        var pass = e.getNext().QueryInterface(Components.interfaces.nsIPassword);
        if (pass.host.indexOf(str)==0) {
          var id=parseInt(pass.host.substring(str.length));
          var obj=new WebMailHandler(this);
          this.webMailHandlers.push(obj);
          obj.id=id;
          obj.user=pass.user;
          obj.password=pass.password;
        }
      }catch(ex) {}
    }
  }else{
    for(var i=0;i<hosts;i++){
      var logins = this.loginManager.findLogins({},str+i,formSubmitURL, null);
      for each (var o in logins) {
        var obj=new WebMailHandler(this);
        this.webMailHandlers.push(obj);
        obj.id=i;
        obj.user=o.username;
        obj.password=o.password;
      }
    }
  }
  var i;
  for(i=0;i<this.webMailHandlers.length;++i){
    var obj=this.webMailHandlers[i];
    try{
      eval("initHandler"+obj.id+"(this.webMailHandlers["+i+"])");
      if(obj.loginData){
        var post=(obj.loginData[LOGIN_DATA]?obj.loginData[LOGIN_DATA]:"");
        if(obj.loginData[LOGIN_ID]){
          if(post)post+="&";
          post+=obj.loginData[LOGIN_ID]+"="+encodeURIComponent(obj.user);
        }
        if(obj.loginData[LOGIN_PW]){
          if(post)post+="&";
          post+=obj.loginData[LOGIN_PW]+"="+encodeURIComponent(obj.password);
        }
        obj.loginData.splice(1,obj.loginData.length-1);
        obj.loginData[LOGIN_POST]=post;
      }
      obj.stage=obj.initStage;
    }catch(e){
      this.webMailHandlers.splice(i,1);//not loaded
      --i;
    }
  }
  this.webMailHandlers.sort(this.sortFunc);
  this.buildTable(aCheckNow);
}
WebMailNotifier.prototype.sortFunc = function(a,b) {
  if(a.id<b.id)return -1;
  else if(a.id==b.id){
    if(a.user<b.user)return -1;
    else if(a.user==b.user)return 0;
    else return 1;
  }else return 1;
}
WebMailNotifier.prototype.buildTable = function(aCheckNow) {
  this.hdlTable=[];
  var defaults=[];
  var hosts=this.prefBranch.getIntPref("hosts");
  for(var i=0;i<hosts;i++){
    this.hdlTable.push([]);
    try{
        defaults[i]=this.prefBranch.getCharPref("webmail"+i+".default");
    }catch(e){}
  }
  for each(var o in this.webMailHandlers){
    try{
      o.enabled=this.prefBranch.getBoolPref("webmail"+o.id+"."+o.user+".enabled");
    }catch(e){}
    if(o.enabled){
      if(o.user==defaults[o.id])this.hdlTable[o.id].splice(0,0,o);//default account inserted into index 0
      else this.hdlTable[o.id].push(o);
    }
  }
  for(var i=0;i<this.hdlTable.length;i++){
    var o=this.hdlTable[i];
    if(o.length>=1){
      var prefStr="webmail"+i+".default";
      var t=null;
      try{
        t=this.prefBranch.getCharPref(prefStr);
      }catch(e){}
      if(t==null)this.prefBranch.setCharPref(prefStr,o[0].user);
    }
  }
  if(aCheckNow){
    this.setState(nsIWebMailNotifier.ST_RESET,"",0);
    this.checkMails();
  }
}
WebMailNotifier.prototype.isDefault = function(id,user) {
  return this.hdlTable[id].length==1||(this.hdlTable[id].length>1&&this.hdlTable[id][0].user==user);
}
WebMailNotifier.prototype.isOpenable = function(id,user) {
  if(this.hdlTable[id].length==1)return true;
  else if(this.hdlTable[id].length==0)return false;
  var ar=this.hdlTable[id];
  for(var i=0;i<ar.length;i++){
    if(ar[i].user==user)break;
    if(ar[i].mailCount>0)return false;
  }
  return true;
}
WebMailNotifier.prototype.isMultiple = function(id) {
  return this.hdlTable[id].length>1;
}
WebMailNotifier.prototype.getHostName = function(id) {
  if(this.hdlTable[id]&&this.hdlTable[id].length>0){
    return this.hdlTable[id][0].name;
  }else{
    var o=new Object();
    try{
      eval("initHandler"+id+"(o)");
    }catch(e){
      return null;
    };
    return o.name;
  }
}
WebMailNotifier.prototype.getAccountsNumber = function() {
  return this.webMailHandlers.length;
}
WebMailNotifier.prototype.getAccountInfo = function(count,aIndex) {
  var obj=this.webMailHandlers[aIndex];
  var ar=[obj.id,obj.user,obj.password];
  count.value=ar.length;
  return ar;
}
WebMailNotifier.prototype.isAccountEnabled = function(aIndex) {
  return this.webMailHandlers[aIndex].enabled;
}
WebMailNotifier.prototype.getIntValue = function(name) {
  return this[name];
}
WebMailNotifier.prototype.setIntValue = function(name,val) {
  this[name]=val;
}

WebMailNotifier.prototype.addListener = function(aListener) {
  this.listeners.push(aListener);
  if (!this.timer){//first listener
    this.checkMails();
    this.setTimer();
  }else{
    for each(var i in this.webMailHandlers){
      var count=this.calcCount(i.id,i.user,i.mailCount);
      aListener.onStateChange(i.id,i.user,count);
    }
  }
}
WebMailNotifier.prototype.removeListener = function(aListener){
  for(var i=0;i<this.listeners.length;i++){
    if(this.listeners[i]==aListener){
        this.listeners.splice(i,1);
        break;
    }
  }
}
WebMailNotifier.prototype.checkMail = function(aID,aUser) {
  var o=this.getHandler(aID,aUser);
  if(o.started){
dout("started "+aID+" "+aUser);
    return;
  }
  o.runCount=0;
  o.checkMail();
}
WebMailNotifier.prototype.checkMails = function(isTimer) {
  for each(var o in this.webMailHandlers){
    if(!isTimer)o.runCount=0;
    o.checkMail();
  }
}
WebMailNotifier.prototype.getHandler = function(aID,aUser) {
  for each(var i in this.webMailHandlers){
    if(i.id==aID&&i.user==aUser)return i;
  }
  return null;
}
WebMailNotifier.prototype.getMailURL = function(aID,aUser) {
  var obj=this.getHandler(aID,aUser);
  if(obj){
    if(this.autoLogin)obj.cookieManager.setCookieToBrowser();
    if(this.newMailsOnly){
      this.prefBranch.setIntPref("webmail"+aID+"."+aUser+".count",obj.mailCount);
      this.setState(aID,aUser,obj.mailCount);
    }
    return obj.getMailURL();
  }else return null;
}
WebMailNotifier.prototype.calcCount = function(aID,aUser,aCount) {
  if(this.newMailsOnly){
    var prefStr="webmail"+aID+"."+aUser+".count";
    var count=null;
    try{
      count=this.prefBranch.getIntPref(prefStr);
    }catch(e){}
    if(count==null){
      this.prefBranch.setIntPref(prefStr,0);
      count=0;
    }
    if(aCount>=0){
      if(aCount>=count)aCount-=count;
      else{
        this.prefBranch.setIntPref(prefStr,aCount);
        aCount=0;
      }
    }
  }
  return aCount;
}
WebMailNotifier.prototype.setState = function(aID,aUser,aCount) {
  if(aID>=0){
    var o=this.getHandler(aID,aUser);
    o.started=false;
    aCount=this.calcCount(aID,aUser,aCount);
    var finished=true;
    for each(o in this.webMailHandlers){
      if(o.started){
        finished=false;
        break;
      }
    }
    if(finished){
      var total=0;
      for each(o in this.webMailHandlers){
        if(o.enabled&&o.mailCount>0){
          total+=this.calcCount(o.id,o.user,o.mailCount);
        }
      }
      if(total>this.countTotal)this.notification(total);
      this.countTotal=total;
//dout("all done");
    }
  }
  for(var i=0;i<this.listeners.length;i++){
    if(this.listeners[i]){
      this.listeners[i].onStateChange(aID,aUser,aCount);
    }
  }
}
WebMailNotifier.prototype.notification = function(total) {
  if(this.prefBranch.getBoolPref("mailAlert")){
    if(this.listeners.length>0){
      try{
        //check alert service
        //doesn't supported in Mac OSX
        var obj = Components.classes["@mozilla.org/alerts-service;1"]
                                       .getService(Components.interfaces.nsIAlertsService);
        this.listeners[0].showAlertNotification(total);
      }catch(e){}
    }
  }
  if(this.prefBranch.getBoolPref("mailSound")){
    if (!this.mSound)this.mSound = Components.classes["@mozilla.org/sound;1"].createInstance(Components.interfaces.nsISound);

    var soundLocation = this.prefBranch.getBoolPref("customSound")?
              this.prefBranch.getCharPref("soundUrl") : "_moz_mailbeep";
    if (soundLocation.indexOf("file://") == -1){
      this.mSound.playSystemSound(soundLocation);
    }else{
      var ioService = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);
      this.mSound.play(ioService.newURI(soundLocation, null, null));
    }
  }
/*

  var obj = Components.classes["@mozilla.org/alerts-service;1"]
                                 .getService(Components.interfaces.nsIAlertsService);
  obj.showAlertNotification("chrome://wm-notifier/content/wmn-icon.png",
                                         "WebMail Notifier","You have new mails", true, "", this);
*/
}

/***********************************************************
timer
***********************************************************/
WebMailNotifier.prototype.setTimer = function() {
  if (this.timer) {
    this.timer.cancel();
  } else {
    this.timer = Components.classes["@mozilla.org/timer;1"]
                  .createInstance(Components.interfaces.nsITimer);
  }
  if(this.timerDelay>0){
    try{
      this.timer.initWithCallback(this,
                                  this.timerDelay,
                                  Components.interfaces.TYPE_ONE_SHOT);
    }catch(e){}
  }
}

WebMailNotifier.prototype.notify = function(aTimer) {
  this.checkMails(true);
  this.setTimer();
}
/***********************************************************
load html
***********************************************************/
WebMailNotifier.prototype.getHtml = function(aHandler,aURL,aPostData,aHeaders) {
  var ioService = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);
  var uri = ioService.newURI(aURL, null, null);
  var channel = ioService.newChannelFromURI(uri);
  var httpChannel = channel.QueryInterface(Components.interfaces.nsIHttpChannel);
  if(aHeaders){
    for(var t in aHeaders){
      httpChannel.setRequestHeader(t,aHeaders[t],false);
    }
  }
  if (aPostData) {
    var uploadStream = Components.classes["@mozilla.org/io/string-input-stream;1"]
                        .createInstance(Components.interfaces.nsIStringInputStream);
    uploadStream.setData(aPostData, aPostData.length);
    var uploadChannel = channel.QueryInterface(Components.interfaces.nsIUploadChannel);
    uploadChannel.setUploadStream(uploadStream, "application/x-www-form-urlencoded", -1);
    httpChannel.requestMethod = "POST";
  }
  aHandler.channel=channel;
  channel.notificationCallbacks = aHandler;
  channel.asyncOpen(aHandler,httpChannel);
}
/***********************************************************
preference observer
***********************************************************/
WebMailNotifier.prototype.observe=function(aSubject, aTopic, aData)
{
  switch(aTopic){
  case "nsPref:changed":
    switch (aData) {
    case "updateInterval":
      this.timerDelay=this.prefBranch.getIntPref(aData)*60000;
      this.setTimer();
      break;
    case "newMailsOnly":
      this.newMailsOnly=this.prefBranch.getBoolPref("newMailsOnly");
      for each(var o in this.webMailHandlers){
        if(o.enabled){
          this.prefBranch.setIntPref("webmail"+o.id+"."+o.user+".count",0);
          this.setState(o.id,o.user,o.mailCount);
        }
      }
      break;
    case "showStatusbarIcon":
      this.setState(nsIWebMailNotifier.ST_SHOWICON,"",this.prefBranch.getBoolPref("showStatusbarIcon"));
      break;
    }
    break;
  case "http-on-modify-request":
    for each(var i in this.webMailHandlers){
      if(i.channel==aSubject){
        i.onRequest(aSubject.QueryInterface(Components.interfaces.nsIHttpChannel));
        break;
      }
    }
    break;
  case "http-on-examine-response":
    for each(var i in this.webMailHandlers){
      if(i.channel==aSubject){
        i.onResponse(aSubject.QueryInterface(Components.interfaces.nsIHttpChannel));
        break;
      }
    }
    break;
  }
}
/***********************************************************
password manager
***********************************************************/
WebMailNotifier.prototype.getPassword=function(host,user){
  if(this.passwordManager){
    var e = this.passwordManager.enumerator;
    while (e.hasMoreElements()) {
      try {
        var pass = e.getNext().QueryInterface(Components.interfaces.nsIPassword);
        if (pass.host == host&&pass.user == user) {
           return pass.password;
        }
      }catch(ex) {}
    }
  }else{
    var logins = this.loginManager.findLogins({},host,formSubmitURL, null);
    for each (var o in logins) {
      if (o.username==user) {
        return o.password;
      }
    }
  }
  return null;
}
WebMailNotifier.prototype.setPassword=function(host,user,passwd){
  if(this.passwordManager){
    try{
      this.passwordManager.removeUser(host,user);
    }catch(e){}
    this.passwordManager.addUser(host,user,passwd);
  }else{
    var nsLoginInfo = new Components.Constructor("@mozilla.org/login-manager/loginInfo;1",
                                                 Components.interfaces.nsILoginInfo,
                                                 "init");
    var loginInfo = new nsLoginInfo(host, formSubmitURL, null, user, passwd, "", "");

    var logins = this.loginManager.findLogins({},host,formSubmitURL, null);
    for each (var o in logins) {
      if (o.username==user) {
        this.loginManager.modifyLogin(o,loginInfo);
        return;
      }
    }
    this.loginManager.addLogin(loginInfo);
  }
}
WebMailNotifier.prototype.removePassword=function(host,user){
  if(this.passwordManager){
    try{
      this.passwordManager.removeUser(host,user);
    }catch(e){}
  }else{
    var logins = this.loginManager.findLogins({},host,formSubmitURL, null);
    for each (var o in logins) {
      if (o.username==user) {
        this.loginManager.removeLogin(o);
        break;
      }
    }
  }
}

WebMailNotifier.prototype.loadFile=function(name){
  var str="";
  try {
    var file;
    if(typeof(name)=="string"){
      file = Components.classes["@mozilla.org/file/directory_service;1"]
                         .getService(Components.interfaces.nsIProperties)
                         .get("ProfD", Components.interfaces.nsIFile);
      name=name.split("/");
      for each(var o in name)file.append(o);
    }else file=name;
    var fis = Components.classes["@mozilla.org/network/file-input-stream;1"]
                          .createInstance(Components.interfaces.nsIFileInputStream);
    fis.init(file, 0x01, 0664, 0);
    var is = Components.classes["@mozilla.org/intl/converter-input-stream;1"]
                       .createInstance(Components.interfaces.nsIConverterInputStream);
    is.init(fis, "UTF-8", 1024, Components.interfaces.nsIConverterInputStream.DEFAULT_REPLACEMENT_CHARACTER);

    var s = {};
    while (is.readString(4096, s) != 0) {
      str+=s.value;
    }
    is.close();
    fis.close();
  }catch(e){}
  return str;
}

WebMailNotifier.prototype.saveFile=function(name,str,append){
  var file = Components.classes["@mozilla.org/file/directory_service;1"]
                     .getService(Components.interfaces.nsIProperties)
                     .get("ProfD", Components.interfaces.nsIFile);
  name=name.split("/");
  for(var i in name){
    file.append(name[i]);
    if(i<name.length-1){
      if(!file.exists()){
        file.create(Components.interfaces.nsIFile.DIRECTORY_TYPE,0755);
      }
    }
  }
  try{
    var fos = Components.classes["@mozilla.org/network/file-output-stream;1"]
                          .createInstance(Components.interfaces.nsIFileOutputStream);
    fos.init(file, 0x02 | 0x08 |(append?0x10:0x20), 0664, 0);
    var os = Components.classes["@mozilla.org/intl/converter-output-stream;1"]
                       .createInstance(Components.interfaces.nsIConverterOutputStream);
    os.init(fos, "UTF-8", 0, 0x0000);
    os.writeString(str);
    os.close();
    fos.close();
  }catch(e){
dout(e);
  }
  return file;
}
WebMailNotifier.prototype.getAvailableID=function(){
  var n=this.loaded.length;
  for(var i=0;i<this.loaded.length;i++){
    if(!this.loaded[i]){
      n=i;
      break;
    }
  }
  if(n==this.loaded.length){//increase hosts
    this.prefBranch.setIntPref("hosts",n+1);
  }
  this.loaded[n]=true;
  return n;
}
WebMailNotifier.prototype.removeHost=function(aHostID){
  try{
    this.prefBranch.deleteBranch("webmail"+aHostID);
    var prefService = Components.classes["@mozilla.org/preferences-service;1"]
                              .getService(Components.interfaces.nsIPrefService);
    prefService.savePrefFile(null);
  }catch(e){
dout(e);
  }
  for each(var o in this.webMailHandlers){
    if(o.id==aHostID){
      this.removePassword("chrome://wm-notifier/webmail"+o.id,o.user);
    }
  }
  eval("initHandler"+aHostID+"=null");
  this.loaded[aHostID]=false;
}
WebMailNotifier.prototype.loadScript=function(name,aHostID){
  try{
    var file = Components.classes["@mozilla.org/file/directory_service;1"]
                       .getService(Components.interfaces.nsIProperties)
                       .get("ProfD", Components.interfaces.nsIFile);
    file.append("wmn");
    file.append(name);

    var ios = Components.classes["@mozilla.org/network/io-service;1"]
                        .getService(Components.interfaces.nsIIOService);
    var fileHandler = ios.getProtocolHandler("file")
                         .QueryInterface(Components.interfaces.nsIFileProtocolHandler);
    var url=fileHandler.getURLSpecFromFile(file);
    var loader =  Components.classes["@mozilla.org/moz/jssubscript-loader;1"]
                    .getService(Components.interfaces.mozIJSSubScriptLoader);
    loader.loadSubScript(url);
    if(aHostID>=0)this.loaded[aHostID]=true;
  }catch(e){
dout(e);
  }
}
/***********************************************************
class factory

This object is a member of the global-scope Components.classes.
It is keyed off of the contract ID.
***********************************************************/
var WebMailNotifierFactory = {
  createInstance: function (aOuter, aIID)
  {
    if (aOuter != null)
      throw Components.results.NS_ERROR_NO_AGGREGATION;
    return (new WebMailNotifier()).QueryInterface(aIID);
  }
};

/***********************************************************
module definition (xpcom registration)
***********************************************************/
var WebMailNotifierModule = {
  _firstTime: true,
  registerSelf: function(aCompMgr, aFileSpec, aLocation, aType)
  {
    aCompMgr = aCompMgr.
        QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.registerFactoryLocation(CLASS_ID, CLASS_NAME,
        CONTRACT_ID, aFileSpec, aLocation, aType);
  },

  unregisterSelf: function(aCompMgr, aLocation, aType)
  {
    aCompMgr = aCompMgr.
        QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.unregisterFactoryLocation(CLASS_ID, aLocation);
  },

  getClassObject: function(aCompMgr, aCID, aIID)
  {
    if (!aIID.equals(Components.interfaces.nsIFactory))
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;

    if (aCID.equals(CLASS_ID))
      return WebMailNotifierFactory;

    throw Components.results.NS_ERROR_NO_INTERFACE;
  },
  canUnload: function(aCompMgr) { return true; }
};

/***********************************************************
module initialization

When the application registers the component, this function
is called.
***********************************************************/
function NSGetModule(aCompMgr, aFileSpec) { return WebMailNotifierModule; }

/***********************************************************
CookieManager definition
***********************************************************/
function CookieInfo(){
}
CookieInfo.prototype={
  getDomain:function(){
    if(this.domain){
      if(this.domain.charAt(0)!=".")return "."+this.domain;
      else return this.domain;
    }
    return this.uri.host;
  },
  getPath:function(){
    if(this.path)return this.path;
    var path=this.uri.path;
    var i=path.lastIndexOf("/");
    if(i==-1)return path;
    else return path.substring(0,i+1);
  },
  toString:function(){
      var s=this.name+"="+this.value;
      for(var p in this){
        if(typeof this[p] == "function")continue;
        if(p!="name"&&p!="value"&&p!="uri"&&p!="httponly"&&p!="secure"){
          s+="; "+p+"="+this[p];
        }
      }
    return s;
  }
}
function CookieManager(){
  this.cookies=[];
}
CookieManager.prototype={
  addCookies: function(aURI,aCookie){
    aCookie=aCookie.split("\n");
    for each(var o in aCookie){
      var val=new CookieInfo();
      var ar=o.split(";");
      var t=ar[0].match(/(\S+?)=(.*)/);

      val["name"]=t[1];
      val["value"]=t[2];

      for(var i=1;i<ar.length;i++){
        if(ar[i]){
          var t=ar[i].replace( /^\s+/g,"");
          var f=t.match(/(\S+?)=(.*)/);
          if(f)val[f[1].toLowerCase()]=f[2];
          else val[t.toLowerCase()]="";
        }
      }
      val.uri=aURI;
      this._insert(val);
//if(val.getPath()!=val.path)dout(val.getDomain()+" "+val.getPath()+" "+val.domain+" "+val.path);
    }
  },
  _insert: function(val){
    for(var i in this.cookies){
      var o=this.cookies[i];
      if(o.name==val.name&&o.getDomain()==val.getDomain()&&o.getPath()==val.getPath()){
        this.cookies[i]=val;
        return;
      }
    }
    this.cookies.push(val);
  },
  findCookie:function(urispec,name){
    for each(var o in this.cookies){
      if(o.uri.spec==urispec&&o.name==name){
        return o.toString();
      }
    }
    return null;
  },
  getCookie: function(aURI){
    var str="";
    for each(var o in this.cookies){
      var domain=o.getDomain();
      if(domain.charAt(0)==".")domain=domain.substring(1);
      if(aURI.host.lastIndexOf(domain)==-1)continue;
      if(aURI.path.indexOf(o.getPath())!=0)continue;
      var exp=o.expires;
      if(exp){
        exp=exp.replace(/-(.+?)-(\d{2}) /," $1 20$2 ").replace(/-/g," ");
        if(Date.parse(exp)<new Date()){
          continue;
        }
      }
      var s=o.name+"="+o.value;
      if(str)str+="; "+s;
      else str=s;
    }
    return str;
  },
  setCookieToBrowser: function(){
    var cs = Components.classes["@mozilla.org/cookieService;1"]
                  .getService(Components.interfaces.nsICookieService);

    var cm = Components.classes["@mozilla.org/cookiemanager;1"]
              .getService(Components.interfaces.nsICookieManager2);
    for each(var o in this.cookies){
      cm.remove(o.getDomain(),o.name,o.getPath(),false);//can't set some cookies if they already exists.
      cs.setCookieString(o.uri,null,o.toString(), null);
    }
  },
  clear: function(){
    this.cookies=[];
  }
}
/***********************************************************
WebMailHandler definition
***********************************************************/
function WebMailHandler(webMailNotifier){
  this.webMailNotifier=webMailNotifier;
  this.cookieManager=new CookieManager();
  this.loggedIn=false;
}
WebMailHandler.prototype={
  charSet: null,
  data: "",
  stage: 0,
  initStage: ST_LOGIN,
  mailCount:-1,
  user: null,
  password: null,
  started:false,
  // nsIChannelEventSink
  onChannelRedirect: function (aOldChannel, aNewChannel, aFlags) {
    if(this.channel==aOldChannel)this.channel = aNewChannel;
  },
  // nsIInterfaceRequestor
  getInterface: function (aIID) {
    try {
      return this.QueryInterface(aIID);
    } catch (e) {
//dout("getInterface "+aIID);
      throw Components.results.NS_NOINTERFACE;
    }
  },
  // nsIProgressEventSink
  onProgress : function (aRequest, aContext, aProgress, aProgressMax) { },
  onStatus : function (aRequest, aContext, aStatus, aStatusArg) { },
  // nsIHttpEventSink
  onRedirect : function (aOldChannel, aNewChannel) { },
  // nsIAuthPromptProvider
  getAuthPrompt : function(aReason){ return null; },
  QueryInterface : function(aIID) {
    if (aIID.equals(Components.interfaces.nsISupports) ||
        aIID.equals(Components.interfaces.nsIInterfaceRequestor) ||
        aIID.equals(Components.interfaces.nsIChannelEventSink) ||
        aIID.equals(Components.interfaces.nsIProgressEventSink) ||
        aIID.equals(Components.interfaces.nsIHttpEventSink) ||
        aIID.equals(Components.interfaces.nsIStreamListener)||
        aIID.equals(Components.interfaces.nsIAuthPromptProvider)||
        aIID.equals(Components.interfaces.nsIPrompt)|| //not used but required in ff 2.0
        aIID.equals(Components.interfaces.nsIDocShell)|| //required in ff 3.0
        aIID.equals(Components.interfaces.nsIAuthPrompt))
      return this;
//dout("QueryInterface "+aIID);
    throw Components.results.NS_NOINTERFACE;
  },
  onStartRequest : function (aRequest, aContext) {
    this.data = "";
  },
  onDataAvailable : function (aRequest, aContext, aStream, aSourceOffset, aLength){
	  var is = Components.classes["@mozilla.org/scriptableinputstream;1"]
            .createInstance(Components.interfaces.nsIScriptableInputStream);
  	is.init(aStream);
  	this.data += is.read(aLength);
  },
  onStopRequest : function (aRequest, aContext, aStatus) {
    var charSet=null;
    var httpChannel=aRequest.QueryInterface(Components.interfaces.nsIHttpChannel);
    if(this.charSet)charSet=this.charSet;
    else{
      var fnd=this.data.match(/<meta.+?\".+?charset=(\S+?)\s*?;?\s*?\"\s*?\x2f?>/);
      if(fnd)charSet=fnd[1];
      else{
        try{
          if(httpChannel.contentCharset)charSet=httpChannel.contentCharset;
        }catch(e){}
      }
    }
    if(charSet){
      var uniConv = Components.classes["@mozilla.org/intl/scriptableunicodeconverter"]
      								.createInstance(Components.interfaces.nsIScriptableUnicodeConverter);
      uniConv.charset = charSet;
      try{
        this.data=uniConv.ConvertToUnicode(this.data);
      }catch(e){}
    }
    if(!this.process(aContext.QueryInterface(Components.interfaces.nsIHttpChannel), this.data))++this.stage;
  },
  process : function (aHttpChannel, aData){
    switch(this.stage){
    case ST_LOGIN:
      this.webMailNotifier.getHtml(this,this.loginData[LOGIN_URL],this.loginData[LOGIN_POST]);
      break;
    case ST_LOGIN_RES:
      this.stage=ST_DATA;
    case ST_DATA:
      this.webMailNotifier.getHtml(this,this.dataURL);
      break;
    case ST_DATA_RES:
      this.mailCount=this.getData(this.data);
      this.stage=this.mailCount<0?this.initStage:ST_DATA;
      this.webMailNotifier.setState(this.id,this.user,this.mailCount);
      if(this.webMailNotifier.autoLogin){
        if(this.mailCount>=0&&!this.loggedIn&&this.webMailNotifier.isDefault(this.id,this.user)){
          this.loggedIn=true;
          this.cookieManager.setCookieToBrowser();
        }
      }
      return true;
    }
  },
  checkMail : function(){
    if(!this.enabled)return;
    this.started=true;
    if(!this.process(null,""))++this.stage;
  },
  onRequest : function(aHttpChannel){
    if(this.stage==this.initStage){
      aHttpChannel.setRequestHeader("Cookie", "", false);//try clean login
      this.cookieManager.clear();
    }else{
      var ck=this.cookieManager.getCookie(aHttpChannel.URI);
      aHttpChannel.setRequestHeader("Cookie",ck, false);
    }
  },
  onResponse : function(aHttpChannel){
    try {
      var cookie = aHttpChannel.getResponseHeader("Set-Cookie");
      this.cookieManager.addCookies(aHttpChannel.URI,cookie);
      aHttpChannel.setResponseHeader("Set-Cookie", "", false);
    }catch(e){}
  },
  getMailURL : function(){
    return this.mailURL;
  }
};

////////////////for debug//////////////////////////////////
function dlog(str){
  dout(str);
  var name="wmn-log.txt";
  if(enableDebug==2)this.notFirst=true;
  str=new Date()+"\t"+str.replace(/\r/g,"\\r").replace(/\n/g,"\\n")+"\r\n";
  WebMailNotifier.prototype.saveFile("wmn-log.txt",str,this.notFirst);
}
///////////////////////////////////////////////////////////
